# Copyright (C) 2012-2013 by Łukasz Langa
# Copyright (C) 2020 by Nguyễn Hồng Quân
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from functools import partial


unset = object()
gettext = unset
no_id_given = -255


class ChoicesEntry(int):
    global_id = 0

    def __new__(cls, *args, **kwargs):
        return super(ChoicesEntry, cls).__new__(
            cls, kwargs['id']
        )

    def __init__(self, description, id, name=None):
        self.raw = description
        self.global_id = ChoicesEntry.global_id
        self.name = name
        self.__extra__ = []
        ChoicesEntry.global_id += 1

    @property
    def desc(self):
        if not self.raw:
            return self.raw
        # gettext obscured that way so you can use choices in settings.py
        global gettext
        if gettext is unset:
            from django.utils.translation import gettext
        return gettext(self.raw)

    @property
    def id(self):
        return int(self)

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value.strip('_') if value else value
        self.__raw_name = value

    def __str__(self):
        name = self.name
        return "<{}: {} (id: {})>".format(self.__class__.__name__,
                                          name, self.id)

    def __repr__(self):
        name = self.name
        name = "{!r}".format(name)
        if name[0] in 'bru':
            name = name[2:-1]
        else:
            name = name[1:-1]
        return "<{}: {} (id: {})>".format(self.__class__.__name__,
                                          name, self.id)

    def extra(self, **other):
        """Enables adding custom attributes to choices at declaration time.
        For example::

            class Color(Choices):
                _ = Choices.Choice

                red = _("red").extra(html='#ff0000')
                green = _("green").extra(html='#00ff00')
                blue = _("blue").extra(html='#0000ff')

        Later on you can use the defined attribute directly::

            >>> Color.red.html
            '#ff0000'

        or with choices received using the getters::

            >>> Color.from_name(request.POST['color']).html
            '#00ff00'
        """
        for key, value in other.items():
            self.__extra__.append(key)
            setattr(self, key, value)
        return self


class ChoiceGroup(ChoicesEntry):
    """A group of choices."""

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, id=kwargs.get('id', args[0]))

    def __init__(self, id, description=''):
        super(ChoiceGroup, self).__init__(description, id=id)
        self.choices = []


class Choice(ChoicesEntry):
    """A single choice."""

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, id=kwargs.get('id', no_id_given))

    def __init__(self, description, id=no_id_given, name=None):
        super(Choice, self).__init__(description, id=id, name=name)
        self.group = None

    def __str__(self):
        return self.desc

    def __repr__(self):
        rawval = self.raw
        name = "{!r}".format(self.name)
        if name[0] in 'bru':
            name = name[2:-1]
        else:
            name = name[1:-1]
        rawval = "{!r}".format(rawval)
        if rawval[0] in 'bru':
            rawval = rawval[2:-1]
        else:
            rawval = rawval[1:-1]
        result = "<{}: {} (id: {}, name: {})>".format(self.__class__.__name__,
                                                      rawval, self.id, name)
        return result


def _getter(name, given, returns, found, getter):

    def impl(cls, id, found=lambda id, k, v: False,
             getter=lambda id, k, v: None, fallback=unset):
        """Unless `fallback` is set, raises ValueError if name not present."""
        for k, v in cls.__dict__.items():
            if isinstance(v, ChoicesEntry) and found(id, k, v):
                return getter(id, k, v)
        if fallback is unset:
            raise ValueError("Nothing found for '{}'.".format(id))
        else:
            return fallback

    function = partial(impl, found=found, getter=getter)
    function.__name__ = name
    function.__doc__ = (
        "Choices.{name}({given}, fallback=unset) -> {returns}"
        "\n\nGiven the `{given}`, returns the `{returns}`. {impl_doc}"
        .format(name=name, given=given, returns=returns,
                impl_doc=impl.__doc__)
    )
    return classmethod(function)


class _ChoicesMeta(type):
    def __new__(meta, classname, bases, class_dict):
        groups = []
        values = []
        raw_values = []
        for k, v in class_dict.items():
            if not isinstance(v, ChoicesEntry):
                continue
            v.name = k
            raw_values.append(v)
        raw_values.sort(key=lambda elem: elem.global_id)
        last_choice_id = 0
        group = None
        for choice in raw_values:
            if isinstance(choice, ChoiceGroup):
                last_choice_id = choice.id
                group = choice
                groups.append(group)
            else:
                if choice.id == no_id_given:
                    last_choice_id += 1
                    c = Choice(choice.raw, id=last_choice_id,
                               name=choice.name)
                    d = dict(((k, getattr(choice, k)) for k in
                              choice.__extra__))
                    choice = c.extra(**d)
                last_choice_id = choice.id
                if group is not None:
                    group.choices.append(choice)
                    choice.group = group
                    d = dict(((k, getattr(group, k)) for k in
                              group.__extra__ if k not in choice.__extra__))
                    choice.extra(**d)
                values.append(choice)
                class_dict[choice._ChoicesEntry__raw_name] = choice
        class_dict['__groups__'] = groups
        class_dict['__choices__'] = values
        return type.__new__(meta, classname, bases, class_dict)


class Choices(list, metaclass=_ChoicesMeta):
    def __init__(self, filter=(unset,), item=unset, grouped=False):
        """Creates a list of pairs from the specified Choices class.
        By default, each pair consists of a numeric ID and the translated
        description. If `use_ids` is False, the name of the attribute
        is used instead of the numeric ID.

        If `filter` is specified, it's a set or sequence of attribute
        names that should be included in the list. Note that the numeric
        IDs are the same regardless of the filtering. This is useful
        for predefining a large set of possible values and filtering to
        only the ones which are currently implemented."""
        if not self.__choices__:
            raise ValueError("Choices class declared with no actual "
                             "choice fields.")
        if item is unset:
            def item(choice):
                return (choice.id, choice.desc)
        filter = set(filter)
        if grouped and self.__groups__:
            for group in self.__groups__:
                group_choices = []
                for choice in group.choices:
                    if (choice.name in filter
                            or (unset in filter and isinstance(choice, Choice))):   # NOQA: W503
                        group_choices.append(item(choice))
                if group_choices:
                    self.append((group.desc, tuple(group_choices)))
        else:
            if grouped:
                import warnings
                warnings.warn("Choices class called with grouped=True and no "
                              "actual groups.")
            for choice in self.__choices__:
                if (choice.name in filter
                        or (unset in filter and isinstance(choice, Choice))):  # NOQA: W503
                    self.append(item(choice))

    from_name = _getter(
        "from_name",
        given="name",
        returns="choice object",
        found=lambda id, k, v: k == id,
        getter=lambda id, k, v: v
    )

    id_from_name = _getter(
        "id_from_name",
        given="name",
        returns="id",
        found=lambda id, k, v: k == id,
        getter=lambda id, k, v: v.id
    )

    desc_from_name = _getter(
        "desc_from_name",
        given="name",
        returns="localized description string",
        found=lambda id, k, v: k == id,
        getter=lambda id, k, v: v.desc
    )

    raw_from_name = _getter(
        "raw_from_name",
        given="name",
        returns="raw description string",
        found=lambda id, k, v: k == id,
        getter=lambda id, k, v: v.raw
    )

    from_id = _getter(
        "from_id",
        given="id",
        returns="choice object",
        found=lambda id, k, v: v.id == id,
        getter=lambda id, k, v: v
    )

    name_from_id = _getter(
        "name_from_id",
        given="id",
        returns="attribute name",
        found=lambda id, k, v: v.id == id,
        getter=lambda id, k, v: k
    )

    desc_from_id = _getter(
        "desc_from_id",
        given="id",
        returns="localized description string",
        found=lambda id, k, v: v.id == id,
        getter=lambda id, k, v: v.desc
    )

    raw_from_id = _getter(
        "raw_from_id",
        given="id",
        returns="raw description string",
        found=lambda id, k, v: v.id == id,
        getter=lambda id, k, v: v.raw
    )

    @staticmethod
    def to_ids(func):
        """Converts a sequence of choices to a sequence of choice IDs."""
        def wrapper(self):
            return (elem.id for elem in func(self))
        return wrapper

    @staticmethod
    def to_names(func):
        """Converts a sequence of choices to a sequence of choice names."""
        def wrapper(self):
            return (elem.name for elem in func(self))
        return wrapper

    Choice = Choice
    Group = ChoiceGroup



# Copyright (C) 2012-2013 by Łukasz Langa
# Copyright (C) 2020 by Nguyễn Hồng Quân
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#
# commonly used choices
#

from textwrap import dedent
from functools import partial

from .core import unset, Choices, ChoicesEntry


class Country(Choices):
    """Specifies a set for all countries of the world (as of January 2011),
    including unions, parts of United Kingdom and unrecognized states."""
    _ = Choices.Choice

    COUNTRIES = Choices.Group(0)
    """Officially recognized states, as of January 2011."""

    af = _("Afghanistan")
    al = _("Albania")
    dz = _("Algeria")
    as_ = _("American Samoa")
    ad = _("Andorra")
    ao = _("Angola")
    ai = _("Anguilla")
    aq = _("Antarctica")
    ag = _("Antigua and Barbuda")
    ar = _("Argentina")
    am = _("Armenia")
    aw = _("Aruba")
    au = _("Australia")
    at = _("Austria")
    az = _("Azerbaijan")
    bs = _("Bahamas")
    bh = _("Bahrain")
    bd = _("Bangladesh")
    bb = _("Barbados")
    by = _("Belarus")
    be = _("Belgium")
    bz = _("Belize")
    bj = _("Benin")
    bm = _("Bermuda")
    bt = _("Bhutan")
    bo = _("Bolivia")
    ba = _("Bosnia and Herzegovina")
    bw = _("Botswana")
    br = _("Brazil")
    bn = _("Brunei")
    bg = _("Bulgaria")
    bf = _("Burkina Faso")
    bi = _("Burundi")
    kh = _("Cambodia")
    cm = _("Cameroon")
    ca = _("Canada")
    cv = _("Cape Verde")
    ky = _("Cayman Islands")
    cf = _("Central African Republic")
    td = _("Chad")
    cl = _("Chile")
    cn = _("China")
    co = _("Colombia")
    km = _("Comoros")
    cg = _("Congo Brazzaville")
    cd = _("Congo Kinshasa")
    ck = _("Cook Islands")
    cr = _("Costa Rica")
    ci = _("Cote Divoire")
    hr = _("Croatia")
    cu = _("Cuba")
    cy = _("Cyprus")
    cz = _("Czech Republic")
    dk = _("Denmark")
    dj = _("Djibouti")
    dm = _("Dominica")
    do = _("Dominican Republic")
    ec = _("Ecuador")
    eg = _("Egypt")
    sv = _("El Salvador")
    gq = _("Equatorial Guinea")
    er = _("Eritrea")
    ee = _("Estonia")
    et = _("Ethiopia")
    fo = _("Faroe Islands")
    fj = _("Fiji")
    fi = _("Finland")
    fr = _("France")
    pf = _("French Polynesia")
    ga = _("Gabon")
    gm = _("Gambia")
    ge = _("Georgia")
    de = _("Germany")
    gh = _("Ghana")
    gi = _("Gibraltar")
    gr = _("Greece")
    gd = _("Grenada")
    gu = _("Guam")
    gt = _("Guatemala")
    gw = _("Guinea Bissau")
    gn = _("Guinea")
    gy = _("Guyana")
    ht = _("Haiti")
    hn = _("Honduras")
    hk = _("Hong Kong")
    hu = _("Hungary")
    is_ = _("Iceland")
    in_ = _("India")
    id = _("Indonesia")
    ir = _("Iran")
    iq = _("Iraq")
    ie = _("Ireland")
    il = _("Israel")
    it = _("Italy")
    jm = _("Jamaica")
    jp = _("Japan")
    je = _("Jersey")
    jo = _("Jordan")
    kz = _("Kazakhstan")
    ke = _("Kenya")
    ki = _("Kiribati")
    kw = _("Kuwait")
    kg = _("Kyrgyzstan")
    la = _("Laos")
    lv = _("Latvia")
    lb = _("Lebanon")
    ls = _("Lesotho")
    lr = _("Liberia")
    ly = _("Libya")
    li = _("Liechtenstein")
    lt = _("Lithuania")
    lu = _("Luxembourg")
    mo = _("Macau")
    mk = _("Macedonia")
    mg = _("Madagascar")
    mw = _("Malawi")
    my = _("Malaysia")
    mv = _("Maldives")
    ml = _("Mali")
    mt = _("Malta")
    mh = _("Marshall Islands")
    mr = _("Mauritania")
    mu = _("Mauritius")
    mx = _("Mexico")
    fm = _("Micronesia")
    md = _("Moldova")
    mc = _("Monaco")
    mn = _("Mongolia")
    me = _("Montenegro")
    ms = _("Montserrat")
    ma = _("Morocco")
    mz = _("Mozambique")
    mm = _("Myanmar")
    na = _("Namibia")
    nr = _("Nauru")
    np = _("Nepal")
    an = _("Netherlands Antilles")
    nl = _("Netherlands")
    nz = _("New Zealand")
    ni = _("Nicaragua")
    ne = _("Niger")
    ng = _("Nigeria")
    kp = _("North Korea")
    no = _("Norway")
    om = _("Oman")
    pk = _("Pakistan")
    pw = _("Palau")
    pa = _("Panama")
    pg = _("Papua New Guinea")
    py = _("Paraguay")
    pe = _("Peru")
    ph = _("Philippines")
    pl = _("Poland")
    pt = _("Portugal")
    pr = _("Puerto Rico")
    qa = _("Qatar")
    ro = _("Romania")
    ru = _("Russian Federation")
    rw = _("Rwanda")
    lc = _("Saint Lucia")
    ws = _("Samoa")
    sm = _("San Marino")
    st = _("Sao Tome and Principe")
    sa = _("Saudi Arabia")
    sn = _("Senegal")
    rs = _("Serbia")
    sc = _("Seychelles")
    sl = _("Sierra Leone")
    sg = _("Singapore")
    sk = _("Slovakia")
    si = _("Slovenia")
    sb = _("Solomon Islands")
    so = _("Somalia")
    za = _("South Africa")
    kr = _("South Korea")
    es = _("Spain")
    lk = _("Sri Lanka")
    kn = _("St Kitts and Nevis")
    vc = _("St Vincent and the Grenadines")
    sd = _("Sudan")
    sr = _("Suriname")
    sz = _("Swaziland")
    se = _("Sweden")
    ch = _("Switzerland")
    sy = _("Syria")
    tj = _("Tajikistan")
    tw = _("Taiwan")
    tz = _("Tanzania")
    th = _("Thailand")
    tl = _("Timor Leste")
    tg = _("Togo")
    to = _("Tonga")
    tt = _("Trinidad and Tobago")
    tn = _("Tunisia")
    tr = _("Turkey")
    tm = _("Turkmenistan")
    tc = _("Turks and Caicos Islands")
    tv = _("Tuvalu")
    ug = _("Uganda")
    ua = _("Ukraine")
    ae = _("United Arab Emirates")
    gb = _("United Kingdom")
    us = _("United States of America")
    uy = _("Uruguay")
    uz = _("Uzbekistan")
    vu = _("Vanuatu")
    va = _("Vatican City")
    ve = _("Venezuela")
    vn = _("Viet Nam")
    vg = _("Virgin Islands British")
    vi = _("Virgin Islands US")
    eh = _("Western Sahara")
    ye = _("Yemen")
    zm = _("Zambia")
    zw = _("Zimbabwe")

    UNITED_KINGDOM = Choices.Group(300)
    """Parts of United Kingdom."""

    england = _("England")
    northern_ireland = _("Northern Ireland")
    wales = _("Wales")
    scotland = _("Scotland")

    UNRECOGNIZED_STATES = Choices.Group(600)
    """De facto countries that are not globally recognized."""
    cy_northern = _("Northern Cyprus")
    palestine = _("Palestine")
    somaliland = _("Somaliland")

    UNIONS = Choices.Group(900)
    """Commonly referred unions and associations."""

    african_union = _("African Union")
    arab_league = _("Arab League")
    association_of_southeast_asian_nations = \
        _("Association of Southeast Asian Nations")
    caricom = _("Caricom")
    commonwealth_of_independent_states = \
        _("Commonwealth of Independent States")
    commonwealth_of_nations = _("Commonwealth of Nations")
    european_union = _("European Union")
    islamic_conference = _("Islamic Conference")
    nato = _("NATO")
    olimpic_movement = _("Olimpic Movement")
    opec = _("OPEC")
    red_cross = _("Red Cross")
    united_nations = _("United Nations")


def _language_lookup_getter(overrides, getter):
    def impl(cls, name, getter=lambda choice: choice, fallback=unset):
        """
        If the given `name` has ``-`` characters, they are converted to
        ``_`` for lookup purposes. If no language is found, a more generic
        language lookup is tried (e.g. for ``"pl-pl"`` also ``"pl"`` will be
        attempted) before raising ValueError or returning the fallback value.
        """
        name = name.replace("-", "_")
        name_generic = name.split("_")[0]
        if name_generic == name:
            name_generic = None
        for k, v in cls.__dict__.items():
            exact_match = k == name
            generic_match = name_generic and k == name_generic
            if isinstance(v, ChoicesEntry) and exact_match or generic_match:
                return getter(v)
        if fallback is unset:
            raise ValueError("Nothing found for '{}'.".format(id))
        else:
            return fallback
    function = partial(impl, getter=getter)
    function.__name__ = overrides.__name__
    function.__doc__ = overrides.__doc__ + " " + dedent(impl.__doc__)
    return classmethod(function)


class Language(Choices):
    """Specifies a broad set of languages. Uses a superset of values found in
    Django and Firefox sources."""

    _ = Choices.Choice

    aa = _("Afar")
    ab = _("Abkhazian")
    ae = _("Avestan")
    af = _("Afrikaans")
    ak = _("Akan")
    am = _("Amharic")
    an = _("Aragonese")
    ar = _("Arabic")
    as_ = _("Assamese")
    ast = _("Asturian")
    av = _("Avaric")
    ay = _("Aymara")
    az = _("Azerbaijani")
    ba = _("Bashkir")
    be = _("Belarusian")
    bg = _("Bulgarian")
    bh = _("Bihari")
    bi = _("Bislama")
    bm = _("Bambara")
    bn = _("Bengali")
    bo = _("Tibetan")
    br = _("Breton")
    bs = _("Bosnian")
    ca = _("Catalan")
    ce = _("Chechen")
    ch = _("Chamorro")
    co = _("Corsican")
    cr = _("Cree")
    cs = _("Czech")
    cu = _("Church Slavic")
    cv = _("Chuvash")
    cy = _("Welsh")
    da = _("Danish")
    de = _("German")
    dv = _("Divehi")
    dz = _("Dzongkha")
    ee = _("Ewe")
    el = _("Greek")
    en = _("English")
    en_gb = _("British English")
    en_us = _("American English")
    eo = _("Esperanto")
    es = _("Spanish")
    es_ar = _("Argentinian Spanish")
    et = _("Estonian")
    eu = _("Basque")
    fa = _("Persian")
    ff = _("Fulah")
    fi = _("Finnish")
    fj = _("Fijian")
    fo = _("Faroese")
    fr = _("French")
    fur = _("Friulian")
    fy = _("Frisian")
    ga = _("Irish")
    gd = _("Scots Gaelic")
    gl = _("Galician")
    gn = _("Guarani")
    gu = _("Gujarati")
    gv = _("Manx")
    ha = _("Hausa")
    he = _("Hebrew")
    hi = _("Hindi")
    ho = _("Hiri Motu")
    hr = _("Croatian")
    hsb = _("Upper Sorbian")
    ht = _("Haitian")
    hu = _("Hungarian")
    hy = _("Armenian")
    hz = _("Herero")
    ia = _("Interlingua")
    id = _("Indonesian")
    ie = _("Interlingue")
    ig = _("Igbo")
    ii = _("Sichuan Yi")
    ik = _("Inupiaq")
    io = _("Ido")
    is_ = _("Icelandic")
    it = _("Italian")
    iu = _("Inuktitut")
    ja = _("Japanese")
    jv = _("Javanese")
    ka = _("Georgian")
    kg = _("Kongo")
    ki = _("Kikuyu")
    kj = _("Kuanyama")
    kk = _("Kazakh")
    kl = _("Greenlandic")
    km = _("Khmer")
    kn = _("Kannada")
    ko = _("Korean")
    kok = _("Konkani")
    kr = _("Kanuri")
    ks = _("Kashmiri")
    ku = _("Kurdish")
    kv = _("Komi")
    kw = _("Cornish")
    ky = _("Kirghiz")
    la = _("Latin")
    lb = _("Luxembourgish")
    lg = _("Ganda")
    li = _("Limburgan")
    ln = _("Lingala")
    lo = _("Lao")
    lt = _("Lithuanian")
    lu = _("Luba-Katanga")
    lv = _("Latvian")
    mg = _("Malagasy")
    mh = _("Marshallese")
    mi = _("Maori")
    mk = _("Macedonian")
    ml = _("Malayalam")
    mn = _("Mongolian")
    mr = _("Marathi")
    ms = _("Malay")
    mt = _("Maltese")
    my = _("Burmese")
    na = _("Nauru")
    nb = _("Norwegian Bokm\u00e5l")
    nd = _("Ndebele, North")
    ne = _("Nepali")
    ng = _("Ndonga")
    nl = _("Dutch")
    nn = _("Norwegian Nynorsk")
    no = _("Norwegian")
    nr = _("Ndebele, South")
    nso = _("Sotho, Northern")
    nv = _("Navajo")
    ny = _("Chichewa")
    oc = _("Occitan")
    oj = _("Ojibwa")
    om = _("Oromo")
    or_ = _("Oriya")
    os = _("Ossetian")
    pa = _("Punjabi")
    pi = _("Pali")
    pl = _("Polish")
    ps = _("Pashto")
    pt = _("Portuguese")
    pt_br = _("Brazilian Portugese")
    qu = _("Quechua")
    rm = _("Rhaeto-Romanic")
    rn = _("Kirundi")
    ro = _("Romanian")
    ru = _("Russian")
    rw = _("Kinyarwanda")
    sa = _("Sanskrit")
    sc = _("Sardinian")
    sd = _("Sindhi")
    se = _("Northern Sami")
    sg = _("Sango")
    si = _("Singhalese")
    sk = _("Slovak")
    sl = _("Slovenian")
    sm = _("Samoan")
    sn = _("Shona")
    so = _("Somali")
    sq = _("Albanian")
    sr = _("Serbian")
    sr_latn = _("Serbian Latin")
    ss = _("Siswati")
    st = _("Sotho, Southern")
    su = _("Sundanese")
    sv = _("Swedish")
    sw = _("Swahili")
    ta = _("Tamil")
    te = _("Telugu")
    tg = _("Tajik")
    th = _("Thai")
    ti = _("Tigrinya")
    tig = _("Tigre")
    tk = _("Turkmen")
    tl = _("Tagalog")
    tlh = _("Klingon")
    tn = _("Tswana")
    to = _("Tonga")
    tr = _("Turkish")
    ts = _("Tsonga")
    tt = _("Tatar")
    tw = _("Twi")
    ty = _("Tahitian")
    ug = _("Uighur")
    uk = _("Ukrainian")
    ur = _("Urdu")
    uz = _("Uzbek")
    ve = _("Venda")
    vi = _("Vietnamese")
    vo = _("Volap\u00fck")
    wa = _("Walloon")
    wen = _("Sorbian")
    wo = _("Wolof")
    xh = _("Xhosa")
    yi = _("Yiddish")
    yo = _("Yoruba")
    za = _("Zhuang")
    zh = _("Chinese")
    zh_cn = _("Simplified Chinese")
    zh_tw = _("Traditional Chinese")
    zu = _("Zulu")

    from_name = _language_lookup_getter(
        overrides=Choices.from_name,
        getter=lambda choice: choice)

    id_from_name = _language_lookup_getter(
        overrides=Choices.id_from_name,
        getter=lambda choice: choice.id)

    desc_from_name = _language_lookup_getter(
        overrides=Choices.desc_from_name,
        getter=lambda choice: choice.desc)

    raw_from_name = _language_lookup_getter(
        overrides=Choices.raw_from_name,
        getter=lambda choice: choice.raw)


class Gender(Choices):
    _ = Choices.Choice

    female = _("female")
    male = _("male")
    unspecified = _("unspecified")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dj', 'dj.choicee']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.0']

setup_kwargs = {
    'name': 'dj.choicee',
    'version': '0.9.1',
    'description': 'An enum implementation for Django field choices, supporting group',
    'long_description': '==========\ndj.choicee\n==========\n\n.. image:: https://badgen.net/pypi/v/dj.choicee\n   :target: https://pypi.org/project/dj.choicee\n.. image:: https://travis-ci.org/AgriConnect/dj.choicee.png\n  :target: https://travis-ci.org/AgriConnect/dj.choicee\n\nThis is a much clearer way to specify choices for fields in models and forms.\nA basic example:\n\n.. code-block:: python\n\n    >>> from dj.choicee import Choices\n    >>> class Gender(Choices):\n    ...   _ = Choices.Choice\n    ...\n    ...   male = _("Male")\n    ...   female = _("Female")\n    ...\n    >>> Gender()\n    [(1, \'Male\'), (2, \'Female\')]\n    >>> Gender.male\n    <Choice: male (id: 1)>\n    >>> Gender.female\n    <Choice: female (id: 2)>\n    >>> Gender.male.id\n    1\n    >>> Gender.male.desc\n    \'Male\'\n    >>> Gender.male.raw\n    \'Male\'\n    >>> Gender.male.name\n    \'male\'\n    >>> Gender.from_name("male")\n    <Choice: male (id: 1)>\n    >>> Gender.id_from_name("male")\n    1\n    >>> Gender.raw_from_name("male")\n    \'Male\'\n    >>> Gender.desc_from_name("male")\n    \'Male\'\n    >>> Gender.name_from_id(2)\n    \'female\'\n    >>> Gender.name_from_id(3)\n    Traceback (most recent call last):\n    ...\n    ValueError: Nothing found for \'3\'.\n    >>> Gender.from_name("perez")\n    Traceback (most recent call last):\n    ...\n    ValueError: Nothing found for \'perez\'.\n\nYou define a class of choices, specifying each choice as a class attribute.\nThose attributes are ``int`` subclasses, numbered automatically starting with\n1.  The class provides several features which support the DRY principle:\n\n * An object instantiated from the choices class is basically a list of ``(id,\n   localized_description)`` pairs straight for consumption by Django.\n\n * Each attribute defined can be retrieved directly from the class.\n\n * Metadata (e.g. attribute name, raw and localized description, numeric ID) of\n   each attribute is accessible.\n\n * Choices which are suffixed by ``_`` to avoid clashing with Python keywords\n   have this suffix automatically removed in their ``.name`` attributes\n\n * Lookup functions are available to help getting attributes or their metadata.\n\n..  note::\n    The ``_ = Choices.Choice`` trick makes it possible for ``django-admin.py\n    makemessages`` to find each choice description and include it in ``.po``\n    files for localization.  It masks ``gettext`` only in the scope of the class so\n    the rest of the module can safely use ``gettext`` or ``gettext_lazy``.  Having to\n    specify ``_`` each time is not a particularly pretty solution but it\'s\n    explicit. Suggestions for a better approach are welcome.\n\nGrouping choices\n----------------\n\nOne of the worst problems with choices is their weak extensibility.  For\ninstance, an application defines a group of possible choices like this:\n\n.. code-block:: python\n\n    >>> class License(Choices):\n    ...   _ = Choices.Choice\n    ...\n    ...   gpl = _("GPL")\n    ...   bsd = _("BSD")\n    ...   proprietary = _("Proprietary")\n    ...\n    >>> License()\n    [(1, \'GPL\'), (2, \'BSD\'), (3, \'Proprietary\')]\n\nAll is well until the application goes live and after a while the developer\nwants to include LGPL.  The natural choice would be to add it after ``gpl`` but\nwhen we do that, the indexing would break.  On the other hand, adding the new\nentry at the end of the definition looks ugly and makes the resulting combo\nboxes in the UI sorted in a counter-intuitive way.  Grouping lets us solve this\nproblem by explicitly defining the structure within a class of choices:\n\n.. code-block:: python\n\n    >>> class License(Choices):\n    ...   _ = Choices.Choice\n    ...\n    ...   COPYLEFT = Choices.Group(0)\n    ...   gpl = _("GPL")\n    ...\n    ...   PUBLIC_DOMAIN = Choices.Group(100)\n    ...   bsd = _("BSD")\n    ...\n    ...   OSS = Choices.Group(200)\n    ...   apache2 = _("Apache 2")\n    ...\n    ...   COMMERCIAL = Choices.Group(300)\n    ...   proprietary = _("Proprietary")\n    ...\n    >>> License()\n    [(1, \'GPL\'), (101, \'BSD\'), (201, \'Apache 2\'), (301, \'Proprietary\')]\n\nThis enables the developer to include more licenses of each group later on:\n\n.. code-block:: python\n\n    >>> class License(Choices):\n    ...   _ = Choices.Choice\n    ...\n    ...   COPYLEFT = Choices.Group(0)\n    ...   gpl_any = _("GPL, any")\n    ...   gpl2 = _("GPL 2")\n    ...   gpl3 = _("GPL 3")\n    ...   lgpl = _("LGPL")\n    ...   agpl = _("Affero GPL")\n    ...\n    ...   PUBLIC_DOMAIN = Choices.Group(100)\n    ...   bsd = _("BSD")\n    ...   public_domain = _("Public domain")\n    ...\n    ...   OSS = Choices.Group(200)\n    ...   apache2 = _("Apache 2")\n    ...   mozilla = _("MPL")\n    ...\n    ...   COMMERCIAL = Choices.Group(300)\n    ...   proprietary = _("Proprietary")\n    ...\n    >>> License()\n    [(1, \'GPL, any\'), (2, \'GPL 2\'), (3, \'GPL 3\'), (4, \'LGPL\'),\n     (5, \'Affero GPL\'), (101, \'BSD\'), (102, \'Public domain\'),\n     (201, \'Apache 2\'), (202, \'MPL\'), (301, \'Proprietary\')]\n\nNote the behaviour:\n\n * the developer renamed the GPL choice but its meaning and ID remained stable\n\n * BSD, Apache and proprietary choices have their IDs unchanged\n\n * the resulting class is self-descriptive, readable and extensible\n\nAs a bonus, the explicitly specified groups can be used when needed:\n\n.. code-block:: python\n\n    >>> License.COPYLEFT\n    <ChoiceGroup: COPYLEFT (id: 0)>\n    >>> License.gpl2 in License.COPYLEFT.choices\n    True\n    >>> [(c.id, c.desc) for c in License.COPYLEFT.choices]\n    [(1, \'GPL, any\'), (2, \'GPL 2\'), (3, \'GPL 3\'), (4, \'LGPL\'),\n     (5, \'Affero GPL\')]\n\n``ChoiceField``\n---------------\n\nChoices can be used with generic ``IntegerField`` and ``CharField`` instances.\nWhen you do that though, some minor API deficiencies show up fairly quickly.\nFirst, when you define the field, you have to instanciate the choices class and\nthe default value has to be converted to the proper type explicitly:\n\n.. code-block:: python\n\n    color = models.IntegerField(choices=Color(), default=Color.green.id)\n\nSecond, when getting the attribute back from a model, it has to be converted to\na ``Choice`` instance to do anything interesting with it:\n\n.. code-block:: python\n\n    >>> obj = Model.objects.get(pk=3)\n    >>> obj.color\n    3\n    >>> Color.from_id(obj.color)\n    <Choice: Blue (id: 3, name: blue)>\n\nTo overcome those problems a ``ChoiceField`` is available in the\n``dj.choicee.fields`` package.  It is based on integers on the database level but\nthe API exposes ``Choice`` instances.  This helps both on the definition side:\n\n.. code-block:: python\n\n    color = ChoiceField(choices=Color, default=Color.green)\n\nand on the access side:\n\n.. code-block:: python\n\n    >>> obj = Model.objects.get(pk=3)\n    >>> obj.color\n    <Choice: Blue (id: 3, name: blue)>\n    >>> obj.color = Color.green\n    >>> obj.save()\n    >>> Model.objects.get(pk=3).color\n    <Choice: Green (id: 2, name: green)>\n\nFor rendering forms, the field coerces to integer values.  That also means that\nwherever ``Choice`` instances are accepted, integers are also fine.\n\nAdvanced functionality\n----------------------\n\nFiltering\n~~~~~~~~~\n\nThe developer can specify all possible choices for future use and then filter\nout only the currently applicable values on choices creation:\n\n.. code-block:: python\n\n    >>> class Language(Choices):\n    ...   _ = Choices.Choice\n    ...\n    ...   de = _("German")\n    ...   en = _("English")\n    ...   fr = _("French")\n    ...   pl = _("Polish")\n    ...\n    >>> Language()\n    [(1, \'German\'), (2, \'English\'), (3, \'French\'), (4, \'Polish\')]\n    >>> Language(filter=("en", "pl"))\n    [(2, \'English\'), (4, \'Polish\')]\n\nThis has the great advantage of keeping the IDs and sorting intact.\n\nCustom item format\n~~~~~~~~~~~~~~~~~~\n\nOne can also change how the pairs are constructed by providing a factory\nfunction.  For instance, to use the class of choices defined above for the\n``LANGUAGES`` setting in ``settings.py``, one could specify:\n\n.. code-block:: python\n\n    >>> Language(item=lambda choice: (choice.name, choice.raw))\n    [(\'de\', \'German\'), (\'en\', \'English\'), (\'fr\', \'French\'),\n     (\'pl\', \'Polish\')]\n\nExtra attributes on choices\n~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nEach choice can receive extra arguments using the ``extra()`` method:\n\n.. code-block:: python\n\n    >>> class Python(Choices):\n    ...   _ = Choices.Choice\n    ...\n    ...   cpython = _("CPython").extra(language=\'C\')\n    ...   pypy = _("PyPy").extra(language=\'Python\')\n    ...   jython = _("Jython").extra(language=\'Java\')\n    ...   iron_python = _("IronPython").extra(language=\'C#\')\n\nThis adds a ``language`` attribute to each choice so you can get it back like\nthis:\n\n.. code-block:: python\n\n    >>> Python.jython.language\n    \'Java\'\n\nThis enables polymorphic attribute access later on when using models or forms.\nFor instance, suppose you have a simple model like:\n\n.. code-block:: python\n\n    >>> class Library(models.Model):\n    ...   name = models.CharField(max_length=100)\n    ...   python_kind = models.IntegerField(choices=Python(), default=Python.cpython.id)\n\nIn that case to get the implementation language back yo\'d do:\n\n.. code-block:: python\n\n    >>> library = Library.objects.get(name=\'dj.choicee\')\n    >>> Python.from_id(library.python_kind).language\n    \'C\'\n\nThat frees your user code of any conditionals or dictionaries that depend on the\nstate of the choices class.  If you would add another choice to it, no user code\nneeds to be changed to support it.  This also supports the DRY principle because\nthe choices class becomes the single place where configuration of that kind is\nheld.\n\nExtra attributes on choice groups\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nUnsurprisingly, choice groups can have extra attributes as well.  They are then\ninherited by choices in such a group and can be overriden if necessary.  For\ninstance:\n\n.. code-block:: python\n\n  >>> class ProfileChange(Choices):\n  ...   _ = Choices.Choice\n  ...\n  ...   USER = Choices.Group(0).extra(icon=\'bookkeeping.png\', is_public=True)\n  ...   email = _("e-mail").extra(is_public=False)\n  ...   first_name = _("first name")\n  ...   last_name = _("last name")\n  ...\n  ...   BASIC_INFO = Choices.Group(10).extra(icon=\'bookkeeping.png\', is_public=True)\n  ...   birth_date = _("birth date").extra(icon=\'calendar.png\')\n  ...   gender = _("gender").extra(icon=\'male_female.png\')\n  ...   country = _("country")\n  ...   city = _("city")\n  ...\n  ...   CONTACT_INFO = Choices.Group(20).extra(icon=\'contactbook.png\', is_public=False)\n  ...   skype = _("Skype ID")\n  ...   icq = _("ICQ number")\n  ...   msn = _("MSN login")\n  ...   xfire = _("X-Fire login")\n  ...   irc = _("IRC info").extra(is_public=True)\n\nIn that case proper inheritance takes place:\n\n.. code-block:: python\n\n  >>> ProfileChange.first_name.is_public\n  True\n  >>> ProfileChange.email.is_public\n  False\n  >>> ProfileChange.country.icon\n  \'bookkeeping.png\'\n  >>> ProfileChange.birth_date.icon\n  \'calendar.png\'\n\n\nPredefined choices\n------------------\n\nThere are several classes of choices which are very common in web applications\nso they are provided already: ``Country``, ``Gender`` and ``Language``.\n\nHow do I run the tests?\n-----------------------\n\nInstall `tox <https://pypi.org/project/tox/>`_ then run:\n\n.. code-block:: sh\n\n  $ tox -e py\n\nChange Log\n----------\n\n1.0.0\n~~~~~\n\n* Pure Python 3 (drop Python 2 compatibility).\n* Support only Python3.5+ and Django 2.0+\n* Fork from `dj.choices`_ to new project.\n\nAuthors\n-------\n\n- Original `dj.choices`_ author `Łukasz Langa <mailto:lukasz@langa.pl>`_.\n- Fork and drop legacy Python2 code: `Nguyễn Hồng Quân <ng.hong.quan@gmail.com>`_.\n\n\n.. _dj.choices: https://pypi.org/project/dj.choices/\n',
    'author': 'Nguyễn Hồng Quân',
    'author_email': 'ng.hong.quan@gmail.com',
    'maintainer': 'Nguyễn Hồng Quân',
    'maintainer_email': 'ng.hong.quan@gmail.com',
    'url': 'https://github.com/AgriConnect/dj.choicee',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)
